from typing import Any
import ast
import copy
from literal import PyPoxLiteral
from variables import PyPoxVariables
from expressions import PyPoxExpression
from statements import PyPoxStatements
from control_flow import PyPoxControlFlow
from functions import PyPoxFunctions
from classes import PyPoxClassDef
class PyPoxCompiler(
    PyPoxLiteral,
    PyPoxVariables,
    PyPoxExpression,
    PyPoxStatements,
    PyPoxControlFlow,
    PyPoxFunctions,
    PyPoxClassDef,
):
    
    # helper functions
        
    def compile(self,ast_node:ast.Module | Any) -> list[str] | str:
        
        if not ast_node:
            return ast_node
        
        if isinstance(ast_node,ast.Module):
            return self.compile(ast_node.body)
        
        if isinstance(ast_node,list):
            body = []
            for node in ast_node:
                body.append(self.compile(node))
            if not all(isinstance(x,str) for x in body):
                return self.compile(body)
            return body    
        
        name = ast_node.__class__.__name__.upper()
        
        if name == 'STR':
            return ast_node
        
        return getattr(self,ast_node.__class__.__name__.upper())(ast_node)
    
if __name__ == "__main__":
    pass


from ast import (
    Name,
    Load,
    Store,
    Del,
    Starred,
)
from base import PyPoxBASE

class PyPoxVariables(PyPoxBASE):
    
    def NAME(self,node:Name):
        
        if isinstance(node.ctx,Load):
            return f"{node.id}"
        
        if isinstance(node.ctx,Store):
            if node.id == node.id.upper():
                return f"const {node.id}"
            else:
                return f"let {node.id}"
        if isinstance(node.ctx,Del):
            return f"delete {node.id}"
    
    def STARRED(self,node:Starred):
        raise Exception("Python Starred not available for convertion")


from ast import (
    Assign,
    AnnAssign,
    AugAssign,
    Raise,
    Assert,
    Delete,
    Pass,
    Import,
    ImportFrom,
    alias,
)
from base import PyPoxBASE

class PyPoxStatements(PyPoxBASE):
    
    def ASSIGN(self,node:Assign):
        
        targets:list[str] = [self.compile(x) for x in node.targets] #type: ignore
        value:str = self.compile(node.value) #type: ignore
        
        return f"{','.join(targets)} = {value.replace(';','')};"
    
    def ANNASSIGN(self,node:AnnAssign):
        raise Exception("ANN-ASSIGN NOT SUPPORTED")
    
    def AUGASSIGN(self,node:AugAssign):
        
        operators:dict[str,str] = {
            "ADD":"+=",
            "SUB":"-=",
            "MULT":"*=",
            "DIV":"/=",
            "MOD":"%=",
        }
        target:str = self.compile(node.target) #type: ignore
        value:str = self.compile(node.value) #type: ignore
        op_name:str = operators[node.op.__class__.__name__.upper()]
        return f"{target.split(' ')[1:][0]} {op_name} {value};"
    
    def RAISE(self,node:Raise):
        return ""
    
    def ASSERT(self,node:Raise):
        raise Exception("ASSERT NOT SUPPORTED")
    
    def DELETE(self,node:Raise):
        raise Exception("DELETE NOT SUPPORTED")
    
    def PASS(self,node:Raise):
        return "/* pass */"
    
    def IMPORT(self,node:Raise):
        return ""
    
    def IMPORTFROM(self,node:Raise):
        return ""


from ast import (
    FunctionDef,
    Lambda,
    arguments,
    arg,
    Return,
)

from base import PyPoxBASE

class PyPoxFunctions(PyPoxBASE):
    
    def FUNCTIONDEF(self,node:FunctionDef):
        func_name = self.compile(node.name) #type: ignore
        func_args = self.compile(node.args.args) #type: ignore
        func_body = self.compile(node.body)
        func_return = self.compile(node.returns) #type: ignore
        return f"const {func_name} = ( {','.join(func_args)} ) => {{{''.join(func_body)}}};"
    
    def ARGUMENTS(self,node:arguments):
        return ""
    
    def RETURN(self,node:Return):
        return f"return {self.compile(node.value).replace(';','')};" #type: ignore
    
    def ARG(self,node:arg):
        return f"{node.arg}"



from ast import (
    ExceptHandler,
    If,
    For,
    Try,
    TryStar,
    While,
    Break,
    Continue,
    withitem,
)

from base import PyPoxBASE

class PyPoxControlFlow(PyPoxBASE):
    
    def IF(self,node:If) -> str:
        
        test: str = self.compile(node.test)
        body: str = self.compile(node.body)
        orelse: str = self.compile(node.orelse)
        
        if orelse:
            if isinstance(node.orelse[0],If):
                return "if ( {test} ) {{ {body} }} else {orelse} ".format(
                    test=test.replace(';',''),
                    body=" ".join(body),
                    orelse=" ".join(orelse)
                )
            else:
                return "if ( {test} ) {{ {body} }} else {{ {orelse} }}".format(
                    test=test.replace(';',''),
                    body=" ".join(body),
                    orelse=" ".join(orelse)
                )
        else:
            return "if ( {test} ) {{ {body} }}".format(
                test=test.replace(';',''),
                body=" ".join(body)
            )
        
    def FOR(self,node:For) -> str:
        
        target = self.compile(node.target)
        iter_ = self.compile(node.iter)
        body = self.compile(node.body)
        orelse = self.compile(node.orelse)
        
        if orelse:
            return "for ( {target} of {iter} ) {{ {body} }} {orelse} ".format(
                target=target.replace(';',''),
                iter=iter_.replace(';',''),
                body=''.join(body),
                orelse=''.join(orelse)
            )
        
        return "for ( {target} of {iter} ) {{ {body} }} ".format(
                target=target.replace(';',''),
                iter=iter_.replace(';',''),
                body=''.join(body)
            ) 
        
    def WHILE(self,node:While) -> str:
        
        test:str = self.compile(node.test)
        body:str = self.compile(node.body)
        orelse:str = self.compile(node.orelse) 
        
        if orelse:
            return "while ( {test} ) {{ {body} }} {orelse}".format(
                test=test,
                body=' '.join(body),
                orelse=' '.join(orelse)
            )
        else:
            return "while ( {test} ) {{ {body} }}".format(
                    test=test,
                    body=' '.join(body),
                )
    
    def BREAK(self,node:Break) -> str:
        return "break;"
    
    def CONTINUE(self,node:Continue) -> str:
        return "continue;"
    
    def TRY(self,node:Try):
        try_block = self.compile(node.body)
        handlers = self.compile(node.handlers)
        #orelse = self.compile(node.orelse)
        #final_body = self.compile(node.finalbody)
        
        try_str:str = "try {{{try_block}}}".format(
            try_block=''.join(try_block)
        )
        catch_str = " ".join(handlers)
        
        return f"{try_str} {catch_str}"
        
    def TRYSTAR(self,node:TryStar):
        raise Exception('TRYSTAR NOT SUPPORTED')
    
    def EXCEPTHANDLER(self,node:ExceptHandler):
        catch_type: str = self.compile(node.type) #type: ignore
        catch_name: str = self.compile(node.name) #type: ignore
        catch_body: str = self.compile(node.body)
        
        return f"catch ({catch_type}) {{{''.join(catch_body)}}}"
    
    def WITH(self,node:Try):
        raise Exception('WITH NOT SUPPORTED')
    
    def WITHITEM(self,node:withitem):
        raise Exception('WITH ITEM NOT SUPPORTED')


from ast import (
    Constant,
    FormattedValue,
    JoinedStr,
    List,
    Load,
    Tuple,
    Set,
    Dict,
    parse,
    dump
)
from base import PyPoxBASE

class PyPoxLiteral(PyPoxBASE):
    
    def CONSTANT(self,node:Constant) -> str:
        
        if isinstance(node.value,str):
            return f"'{node.value}'"
        else:
            return f"{node.value}"

    def FORMATTEDVALUE(self,node:FormattedValue) -> str:
        return ""

    def JOINEDSTR(self,node:JoinedStr):
        
        for val in node.values:
            pass

    def LIST(self,node: List) -> str:
        
        if isinstance(node.ctx,Load):
            return f"[{','.join([self.compile(x) for x in node.elts])}]" #type: ignore
        
        return ""

    def TUPLE(self,node: Tuple) -> str:
        
        if isinstance(node.ctx,Load):
            return f"[{','.join([self.compile(x) for x in node.elts])}]" #type: ignore
        
        return ""
    
    def SET(self,node: Set) -> str:
        
        return f"new Set([{','.join([self.compile(x) for x in node.elts])}])" #type: ignore

    def DICT(self,node: Dict) -> str:
        
        pairs:list[str] = []
        
        for key,val in zip(node.keys,node.values):
            pairs.append(f"{self.compile(key)}:{self.compile(val)}") #type: ignore

        return "{" + ",".join(pairs) + "}"


from ast import (
    Attribute,
    BinOp,
    BoolOp,
    Call,
    Compare,
    DictComp,
    Expression,
    Expr,
    GeneratorExp,
    IfExp,
    ListComp,
    NamedExpr,
    SetComp,
    Slice,
    Subscript,
    UnaryOp,
    keyword,
)

from base import PyPoxBASE

class PyPoxExpression(PyPoxBASE):
    
    def EXPRESSION(self,node:Expression):
        return self.compile(node.body) #type: ignore
        
    def EXPR(self,node:Expr):
        return self.compile(node.value) #type: ignore
    
    def UNARYOP(self,node:UnaryOp):
        
        operators:dict[str,str|Exception] = {
            "UADD":Exception("UADD NOT SUPPORTED"),
            "NOT":"!",
            "USUB":"-",
            "INVERT":"~",
        }
        op_name: str | Exception = operators[node.op.__class__.__name__.upper()]
        
        if isinstance(op_name,Exception):
            raise op_name
        
        return f"{op_name}{self.compile(node.operand)}" #type: ignore

    def BINOP(self,node:BinOp):
        
        left = self.compile(node.left)
        right = self.compile(node.right)
        
        operators:dict[str,str|Exception] = {
            "ADD":"+",
            "SUB":"-",
            "MULT":"*",
            "DIV":"/",
            "FLOORDIV":Exception("FLOOR DIV NOT SUPPORTED"),
            "MOD":"%",
            "POW":"**",
            "LSHIFT":"<<",
            "RSHIFT":">>",
            "BITOR":"|",
            "BITXOR":"^",
            "BITAND":"&",
            "MATMUL":Exception("MAT MUL NOT SUPPORTED"),
        }
        
        op_name: str | Exception = operators[node.op.__class__.__name__.upper()]
        
        if isinstance(op_name,Exception):
            raise op_name
        
        return f"{left} {op_name} {right}"
    
    def BOOLOP(self,node:BoolOp) -> None:
        
        operators:dict[str,str] = {
            "AND":"&&",
            "OR":"||"
        }
        
        op_name = operators[node.op.__class__.__name__.upper()]
        
        return f" {op_name} ".join([self.compile(x) for x in node.values]) # type: ignore
        
    def COMPARE(self,node:Compare):
        
        ops_dict = {
            'EQ':'===',
            'NOTEQ':'!==',
            'LT':'<',
            'LTE':'<=',
            'GT':'>',
            'GTE':'>=',
            'IS':'===',
            'ISNOT':'!==',
            'IN':Exception('Not supported'),
            'NOTIN':Exception('Not supported'),
        }
        
        left = self.compile(node.left)
        ops = [ops_dict[x.__class__.__name__.upper()] for x in node.ops]
        comparators = [self.compile(x) for x in node.comparators]
        
        return f"{left} {''.join([f'{o} {c}' for o,c in zip(ops,comparators)])}"  
    
    def CALL(self,node:Call):
        
        if not isinstance(node,Call):
            raise Exception("Invalid ast node must be AST Call")
        
        func_name:str = self.compile(node.func) #type: ignore
        args:list[str] = [self.compile(x) for x in node.args] #type: ignore
        return f"{func_name.replace(';','')}({','.join([x.replace(';','') for x in args])});"
    
    def KEYWORD(self,node:keyword):
        args = self.compile(node.arg) #type: ignore
        value = self.compile(node.value)
    
    def IFEXP(self,node:IfExp):
        
        if not isinstance(node,IfExp):
            raise Exception("Invalid ast node must be AST IfExp")
        
        test:str = self.compile(node.test) #type: ignore
        body:str = self.compile(node.body) #type: ignore
        orelse:str = self.compile(node.orelse) #type: ignore
        
        return "{test} ? {body} : {orelse}".format(
            test=test.replace(';',''),
            body=body.replace(';',''),
            orelse=orelse.replace(';',''),
        ) 
    
    def ATTRIBUTE(self,node:Attribute):
        
        value = getattr(self,node.value.__class__.__name__.upper())(node.value)
        attr = node.attr
        ctx = node.ctx
        
        return f"{value.replace(';','')}.{attr}"
    
    def NAMEDEXPR(self,node:NamedExpr):
        raise Exception("NAMED EXPR NOT SUPPORTED")
    
    def SUBSCRIPT(self,node:Subscript):
        raise Exception("SUBSCRIPT NOT SUPPORTED")
    
    def SLICE(self,node:Slice):
        raise Exception("SLICE NOT SUPPORTED")
    
    def LISTCOMP(self,node:ListComp):
        raise Exception("LISTCOMP NOT SUPPORTED")
    
    def SETCOMP(self,node:SetComp):
        raise Exception("SETCOMP NOT SUPPORTED")
    
    def GENERATOREXP(self,node:GeneratorExp):
        raise Exception("GENERATOREXP NOT SUPPORTED")
    
    def DICTCOMP(self,node:DictComp):
        raise Exception("DICTCOMP NOT SUPPORTED")


from importlib.metadata import entry_points
from setuptools import setup,find_packages
import os
setup(
    name='pypox',
    version=str(os.getenv('__package_version__')), #type: ignore
    author='Karl Robeck Alferez',
    author_email='karlalferezfx@gmail.com',
    description='Python to javascript transpiler',
    packages=find_packages(),
    entry_points = {
      'console_script':[
          'pypox = pypox.__main__:main'
      ]  
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.11.4',
)


from ast import (
    ClassDef,
)

from base import PyPoxBASE

class PyPoxClassDef(PyPoxBASE):
    
    def CLASSDEF(self,node:ClassDef):
        class_name = self.compile(node.name) #type: ignore
        class_bases = self.compile(node.bases)
        class_body = self.compile(node.body)
        class_methods = [x.split('=')[0].split(' ')[1] for x in class_body]
        
        
        class_body = [old.replace(old.split('=')[0],new) for old,new in zip(class_body,class_methods)]
        
        if class_bases:
            return f"class {class_name} extends {','.join(class_bases)} {{{''.join(class_body)}}};"
        else:
            return f"class {class_name} {{{''.join(class_body)}}};"